@@ -1,5 +1,5 @@
 from __future__ import absolute_import
-from pony.py23compat import PY2, imap, basestring, buffer, int_types
+from pony.py23compat import buffer, int_types
 
 import json
 from decimal import Decimal
@@ -47,10 +47,21 @@ class MySQLTranslator(SQLTranslator):
     dialect = 'MySQL'
     json_path_wildcard_syntax = True
 
+class MySQLValue(Value):
+    __slots__ = []
+    def __str__(self):
+        value = self.value
+        if isinstance(value, timedelta):
+            if value.microseconds:
+                return "INTERVAL '%s' HOUR_MICROSECOND" % timedelta2str(value)
+            return "INTERVAL '%s' HOUR_SECOND" % timedelta2str(value)
+        return Value.__str__(self)
+
 class MySQLBuilder(SQLBuilder):
     dialect = 'MySQL'
+    value_class = MySQLValue
     def CONCAT(builder, *args):
-        return 'concat(',  join(', ', imap(builder, args)), ')'
+        return 'concat(',  join(', ', map(builder, args)), ')'
     def TRIM(builder, expr, chars=None):
         if chars is None: return 'trim(', builder(expr), ')'
         return 'trim(both ', builder(chars), ' from ' ,builder(expr), ')'
@@ -79,21 +90,21 @@ class MySQLBuilder(SQLBuilder):
     def SECOND(builder, expr):
         return 'second(', builder(expr), ')'
     def DATE_ADD(builder, expr, delta):
-        if isinstance(delta, timedelta):
-            return 'DATE_ADD(', builder(expr), ", INTERVAL '", timedelta2str(delta), "' HOUR_SECOND)"
-        return 'ADDTIME(', builder(expr), ', ', builder(delta), ')'
+        if delta[0] == 'VALUE' and isinstance(delta[1], time):
+            return 'ADDTIME(', builder(expr), ', ', builder(delta), ')'
+        return 'ADDDATE(', builder(expr), ', ', builder(delta), ')'
     def DATE_SUB(builder, expr, delta):
-        if isinstance(delta, timedelta):
-            return 'DATE_SUB(', builder(expr), ", INTERVAL '", timedelta2str(delta), "' HOUR_SECOND)"
-        return 'SUBTIME(', builder(expr), ', ', builder(delta), ')'
+        if delta[0] == 'VALUE' and isinstance(delta[1], time):
+            return 'SUBTIME(', builder(expr), ', ', builder(delta), ')'
+        return 'SUBDATE(', builder(expr), ', ', builder(delta), ')'
+    def DATE_DIFF(builder, expr1, expr2):
+        return 'TIMEDIFF(', builder(expr1), ', ', builder(expr2), ')'
     def DATETIME_ADD(builder, expr, delta):
-        if isinstance(delta, timedelta):
-            return 'DATE_ADD(', builder(expr), ", INTERVAL '", timedelta2str(delta), "' HOUR_SECOND)"
-        return 'ADDTIME(', builder(expr), ', ', builder(delta), ')'
+        return builder.DATE_ADD(expr, delta)
     def DATETIME_SUB(builder, expr, delta):
-        if isinstance(delta, timedelta):
-            return 'DATE_SUB(', builder(expr), ", INTERVAL '", timedelta2str(delta), "' HOUR_SECOND)"
-        return 'SUBTIME(', builder(expr), ', ', builder(delta), ')'
+        return builder.DATE_SUB(expr, delta)
+    def DATETIME_DIFF(builder, expr1, expr2):
+        return 'TIMEDIFF(', builder(expr1), ', ', builder(expr2), ')'
     def JSON_QUERY(builder, expr, path):
         path_sql, has_params, has_wildcards = builder.build_json_path(path)
         return 'json_extract(', builder(expr), ', ', path_sql, ')'
@@ -175,7 +186,7 @@ class MySQLJsonConverter(dbapiprovider.JsonConverter):
     NE = 'JSON_NE'
     def init(self, kwargs):
         if self.provider.server_version < (5, 7, 8):
-            version = '.'.join(imap(str, self.provider.server_version))
+            version = '.'.join(map(str, self.provider.server_version))
             raise NotImplementedError("MySQL %s has no JSON support" % version)
 
 class MySQLProvider(DBAPIProvider):
@@ -200,7 +211,7 @@ class MySQLProvider(DBAPIProvider):
     converter_classes = [
         (NoneType, dbapiprovider.NoneConverter),
         (bool, dbapiprovider.BoolConverter),
-        (basestring, MySQLStrConverter),
+        (str, MySQLStrConverter),
         (int_types, dbapiprovider.IntConverter),
         (float, MySQLRealConverter),
         (Decimal, dbapiprovider.DecimalConverter),
@@ -230,19 +241,13 @@ class MySQLProvider(DBAPIProvider):
         cursor.execute('set session group_concat_max_len = 4294967295')
 
     def should_reconnect(provider, exc):
-        return isinstance(exc, mysql_module.OperationalError) and exc.args[0] == 2006
+        return isinstance(exc, mysql_module.OperationalError) and exc.args[0] in (2006, 2013)
 
     def get_pool(provider, *args, **kwargs):
         if 'conv' not in kwargs:
             conv = mysql_converters.conversions.copy()
             if mysql_module_name == 'MySQLdb':
                 conv[FIELD_TYPE.BLOB] = [(FLAG.BINARY, buffer)]
-            else:
-                if PY2:
-                    def encode_buffer(val, encoders=None):
-                        return string_literal(str(val), encoders)
-
-                    conv[buffer] = encode_buffer
 
             def encode_timedelta(val, encoders=None):
                 return string_literal(timedelta2str(val), encoders)
@@ -336,6 +341,6 @@ def str2datetime(s):
     if 19 < len(s) < 26: s += '000000'[:26-len(s)]
     s = s.replace('-', ' ').replace(':', ' ').replace('.', ' ').replace('T', ' ')
     try:
-        return datetime(*imap(int, s.split()))
+        return datetime(*map(int, s.split()))
     except ValueError:
         return None  # for incorrect values like 0000-00-00 00:00:00
